"""A module providing selection strategies for the Kaczmarz algorithm."""

import kaczmarz


class Cyclic(kaczmarz.Base):
    """Cycle through the equations of the system in order, repeatedly.

    Note
    ----
    This class inherits the parameters, methods, and properties of
    :class:`kaczmarz.Base`.

    Parameters
    ----------
    base_args : tuple
        Positional arguments for :class:`kaczmarz.Base` constructor.
    base_kwargs : dict
        Keyword arguments for :class:`kaczmarz.Base` constructor.

    References
    ----------
    1. Kaczmarz, S.: Angenäherte Auflösung von Systemen linearer Gleichungen.
       *Bulletin International de l’Académie Polonaise
       des Sciences et des Lettres.
       Classe des Sciences Mathématiques et Naturelles.
       Série A, Sciences Mathématiques* 35, 335–357 (1937)
    """

    def __init__(self, *base_args, **base_kwargs):
        super().__init__(*base_args, **base_kwargs)
        self.n_rows = self._A.shape[0]
        self.row_index = -1

    def _select_row_index(self, xk):
        """Get the next row index, in cyclic order.

        Parameters
        ----------
        xk : (n,) array
            The current Kaczmarz iterate. Not used for this selection strategy.

        Returns
        -------
        ik : int
            The next row index, in cyclic order.
        """
        self.row_index = (1 + self.row_index) % self.n_rows
        return self.row_index



import numpy as np
import pytest

import kaczmarz


def test_row_indexes(eye23, ones2):
    x = np.zeros(3)
    cyclic = kaczmarz.Cyclic(eye23, ones2)
    assert 0 == cyclic._select_row_index(x)
    assert 1 == cyclic._select_row_index(x)
    assert 0 == cyclic._select_row_index(x)
    assert 1 == cyclic._select_row_index(x)

    x0 = np.zeros(3)
    iterates = kaczmarz.Cyclic.iterates(eye23, ones2, x0)
    iterator = iter(iterates)
    assert [0, 0, 0] == list(next(iterator))
    assert [1, 0, 0] == list(next(iterator))
    assert [1, 1, 0] == list(next(iterator))
    with pytest.raises(StopIteration):
        next(iterator)


def test_solve_identity(eye33, ones3):
    x = kaczmarz.Cyclic.solve(eye33, ones3)
    assert [1, 1, 1] == list(x)


def test_solve_non_orthogonal_matrix():
    Alol = [
        [1, 0, 0],
        [2, 1, 0],
        [1, 2, 1],
        [0, 1, 2],
        [0, 0, 1],
    ]
    A = np.array(Alol)
    row_norms = np.linalg.norm(A, axis=1)
    x_exact = np.ones(3)
    b = A @ x_exact
    tol = 1e-5
    x = kaczmarz.Cyclic.solve(A, b, tol=tol)
    assert np.linalg.norm(A @ (x - x_exact) / row_norms) < tol
